# -*- coding: utf-8 -*-

from __future__ import print_function

import io
import json
import logging
import os
import re
import shutil
import sys
import textwrap
from argparse import (
    Action,
    ArgumentParser,
    RawTextHelpFormatter,
    _SubParsersAction
)
from tempfile import NamedTemporaryFile

from .compat import init, str_type
from .parser import Parser, ParseResult, Writer


def parse_query(raw):
    """Parse query to the libconfuse structure.  The query is passed
    as a JSON array.  It has to have one or more strings followed by
    zero or more integers.  The strings until the last one must have
    values of struct names.  The last one must the the lhs of an
    assignment statement.  Integers are indexes into a list, if the
    rhs of an assignment is a list.  It seems that libconfuse doesn't
    support nested lists, but this parser implementation does.

    :param raw: The unparsed query.
    :type raw: str

    :returns: List of one or more strings followed by zero or more
        integers.
    :rtype: List[Union[str, int]]
    """
    return json.loads(raw)


def _common_arguments(parser, need_value=True, need_match=False):
    if need_match:
        query = parser.add_mutually_exclusive_group(required=True)
        query.add_argument(
            "-m",
            "--match",
            action="append",
            type=str_type(),
            help=textwrap.dedent("""
            Similar to --query, however uses regular expression to
            match the whole path to the value.  This will potentially
            select multiple values.  It is possible to ask to ignore
            case when matching.  Use /pattern/i to do that."""),
        )
    else:
        query = parser

    query.add_argument(
        "-q",
        "--query",
        type=parse_query,
        help=textwrap.dedent("""
        Path to the value that needs to be modified.  The value is
        given as a JSON array. It should contain one or more string
        followed by zero or more integers.  Strings are used to
        recursively select lhs expressions (be those sections or
        assignments), while integers are used to select elements from
        lists."""),
    )
    if need_value:
        parser.add_argument(
            "-e",
            "--value",
            help=textwrap.dedent("""
            New value to use when modifying the document.  Can be
            given using either JSON format or libconfuse format,
            depending on --input-format."""),
        )
    parser.add_argument(
        "infile",
        help=textwrap.dedent("""
        File with libconfuse configuration to read.  To read stdin use
        `-` as a value."""),
    )


# Copied with some modifictions from: <https://gist.github.com/sampsyo/471779>

class AliasedSubParsersAction(_SubParsersAction):
    """ArgumentParser uses the concept of "action" to describe many
    different entities.  In particular, it represents an addition of
    subparsers ad an action (this is also how it later generates help
    info for them).  It is possible to override the ``parser`` action
    (a pre-defined action that is used to create subparsers) with a
    class that implements interface similar to ``_SubParsersAction``.

    The built-in way for generating aliases doesn't group them, which
    makes documentation (specifically, the usage section) harder to
    understand.

    Below is what aliases may look like:

    .. code-block:: console

        command delete <file>

    and

    .. code-block:: console

        command del <file>

    You would have to do something like this:

    .. code-block:: python

        parser = ArgumentParser()
        subparsers = parser.add_subparsers()
        delete_subparser = subparsers.add_parser("delete")
        # define interface for delete_subparser
        del_subparser = subparsers.add_parser("del")
        # copy interface delete_subparser to del_subparser

    It's possible, but inconvenient.  That's why I added this class to
    override the :func:`~argparse._SubParsersAction.add_parser()` to
    allow it to take an extra argument ``aliases`` to reduce the
    amount of duplication.

    """

    class _AliasedPseudoAction(Action):
        def __init__(self, name, aliases, help):
            dest = name
            if aliases:
                dest += "({})".format(','.join(aliases))
            super(_AliasedPseudoAction, self).__init__(
                option_strings=[],
                dest=dest,
                help=help,
            )

    def add_parser(self, name, **kwargs):
        if "aliases" in kwargs:
            aliases = kwargs["aliases"]
            del kwargs["aliases"]
        else:
            aliases = []

        parser = super(
            AliasedSubParsersAction,
            self,
        ).add_parser(name, **kwargs)

        # Make the aliases work.
        for alias in aliases:
            self._name_parser_map[alias] = parser
        # Make the help text reflect them, first removing old help entry.
        if "help" in kwargs:
            help = kwargs.pop("help")
            self._choices_actions.pop()
            pseudo_action = self._AliasedPseudoAction(name, aliases, help)
            self._choices_actions.append(pseudo_action)

        if hasattr(self, "_aliases"):
            paliases = self._aliases
        else:
            paliases = {}
        paliases[name] = aliases
        self._aliases = paliases

        return parser


class AliasHelpFormatter(RawTextHelpFormatter):
    """This class implements help formatter for
    :class:`~argparse.ArgumentParser`.  In particular, it helps
    formatting the aliases by grouping them together."""

    def _metavar_formatter(self, action, default_metavar):
        if action.metavar is not None:
            result = action.metavar
        elif isinstance(action, AliasedSubParsersAction):
            parts = []
            for k, v in action._aliases.items():
                if v:
                    parts.append("({}|{})".format(k, '|'.join(v)))
                else:
                    parts.append(k)
            result = "{" + ",".join(parts) + "}"
        elif action.choices is not None:
            choice_strs = [str(choice) for choice in action.choices]
            result = "{" + ",".join(choice_strs) + "}"
        else:
            result = default_metavar

        def format(tuple_size):
            if isinstance(result, tuple):
                return result
            else:
                return (result,) * tuple_size
        return format


def create_parser():
    """
    Creates parser for command-line options for ``cm-confuse`` utility.

    :returns: Parser for ``cm-confuse`` utility.
    :rtype: ~argparse.ArgumentParser

    .. seealso::

        :doc:`CLI <../cli>`
    """
    parser = ArgumentParser(
        description="Bright Cluster Manager: libconfuse parsing utility",
        formatter_class=AliasHelpFormatter,
        epilog=textwrap.dedent("""
        Some commands in this tool can be abbreviated or have aliases.
        The notation `(x|y|...|z)` in positional arguments section
        groups the availabe abbreviations and aliases.  Some examples
        of how this utility may be used:

            cm-confuse s -m LDAP /cm/local/apps/cmd/etc/cmd.conf

        To get all configuration variables related to LDAP
        (eg. LDAPHost, LDAPUser and so on).

            cm-confuse rm -m '/^DB/i' /cm/local/apps/cmd/etc/cmd.conf

        Deletes all configuration variables whose names start with "DB"
        ignoring case.

            cm-confuse a -q '["AdvancedConfig"]' -e '"CmjobRestartNFSServer=1"'

        Will append `CmjobRestartNFSServer=1' to the list AdvancedConfig.

            cm-confuse u -q '["EventLoggerFile"] -e '"/tmp/events.log"'

        Will replace the value of EventLoggerFile with /tmp/events.log.
        """),
    )
    parser.register("action", "parsers", AliasedSubParsersAction)
    parser.add_argument(
        "-o",
        "--output",
        default=None,
        help=textwrap.dedent("""
        Where to output the result of the query.  To output to stdout
        use `-` as a value. When selecting or removing values, the
        default is to print the result to the stdout.  When replacing
        or appending, the default is to output to the same file as the
        source."""),
    )
    parser.add_argument(
        "-f",
        "--input-format",
        default="json",
        choices=("json",),
        help=textwrap.dedent("""
        Format to use when reading values to be used when patching
        libonconfuse documents."""),
    )
    parser.add_argument(
        "-F",
        "--output-format",
        default="confuse",
        choices=("json", "confuse", "table"),
        help="""Format to use when writing patched documents.""",
    )
    parser.add_argument(
        "-V",
        "--verbosity",
        default="warning",
        choices=(
            "critical",
            "debug",
            "error",
            "fatal",
            "info",
            "warning",
            "warn",
            "notset",
        ),
        help=textwrap.dedent("""
        Verbosity of logging.  Use constants from `logging` package to
        set the level."""),
    )
    actions = parser.add_subparsers()
    actions.required = True

    select = actions.add_parser("select", aliases=("s", "sel", "get"))
    update = actions.add_parser("update", aliases=("u", "up"))
    append = actions.add_parser("append", aliases=("a", "ap"))
    remove = actions.add_parser("remove", aliases=("r", "rm", "rem", "del"))
    select.set_defaults(action="select")
    update.set_defaults(action="update")
    append.set_defaults(action="append")
    remove.set_defaults(action="remove")

    _common_arguments(select, need_value=False, need_match=True)
    _common_arguments(update)
    _common_arguments(append)
    _common_arguments(remove, need_match=True)

    return parser


def in_formatter(fmt):
    """Returns an implementation (a function) for `fmt` s.t. this
    function is able to parse the format in which the value to the
    modification query is given.

    :param fmt: One of the supported formats (``"json"``,
        ``"confuse"``).
    :type fmt: str

    :returns: A function that given a string representation of the
        value will regurn Python value.
    :rtype: Callable[str, any]
    """
    if fmt == "json":
        return json.loads
    if fmt == "confuse":
        return Parser().parse
    raise NotImplementedError("Unrecognized input format: {}".format(fmt))


class SameFile:
    """Creates a callable object used for output when writing the new
    libconfuse contents to the same file.

    :ivar f: The name of the file that will be overwritten with the
        new contents.
    :type f: str
    """

    def __init__(self, f):
        self.f = f

    def __call__(self, formatted):
        with NamedTemporaryFile() as tf:
            tf.write(formatted.encode("utf-8"))
            tf.write(b"\n")
            tf.flush()
            logging.debug("Copying %s to %s", tf.name, self.f)
            shutil.copy(tf.name, self.f)


class OtherFile:
    """Creates a callable object used for output when writing the new
    libconfuse contents to a different file.

    :ivar f: The name of the file that will be overwritten with the
        new contents.
    :type f: str
    """

    def __init__(self, f):
        self.f = f

    def __call__(self, formatted):
        with open(self.f, "w") as w:
            w.write(formatted)
            w.write("\n")


class MultiPattern:

    def __init__(self, patterns):
        self.patterns = tuple(self._compile(p) for p in patterns)

    def _compile(self, p):
        if p.startswith("/") and p.endswith("/i"):
            # It seems to be only reasonable to support ignore case option
            # Maybe later we can support other options, like extended etc.
            return re.compile(p[1:-2], re.IGNORECASE)
        elif (p.startswith("/") and p.endswith("/")):
            return re.compile(p[1:-1])
        elif not (p.startswith("/") or p.endswith("/")):
            return re.compile(p)
        raise SyntaxError("Only ingorecase option is supported: {}".format(p))

    def search(self, string, pos=0, endpos=9223372036854775807):
        for p in self.patterns:
            m = p.search(string, pos, endpos)
            if m:
                return m


def _format_table(value):
    if isinstance(value, ParseResult):
        value = value.data()
    col_a, col_b = ["Name"], ["Value"]
    max_a, max_b = len(col_a[0]), len(col_b[0])
    for k, v in value.items():
        sk = str(k)
        max_a = max(max_a, len(sk))
        max_b = max(max_b, len(repr(v)))
        col_a.append(sk)
        col_b.append(v)
    htpl = "{:<" + str(max_a) + "} | {:>" + str(max_b) + "}"
    tpl = "{:<" + str(max_a) + "} | {!r:>" + str(max_b) + "}"
    sep = "-" * (max_a + max_b + 3)
    lines = [htpl.format(col_a[0], col_b[0])] + [sep] + [
        tpl.format(a, b) for a, b in zip(col_a[1:], col_b[1:])
    ]
    return "\n".join(lines)


def _json_dumps(s):
    if isinstance(s, ParseResult):
        s = s.data()
    return json.dumps(s, indent=2)


def output(value, args):
    """Handles the output from the operation being performed on the
    libconfuse file.

    :param value: The value to output.
    :type value: any

    :param args: The parsed command-line arguments.
    :type args: Namespace
    """
    if args.output == "-":
        handler = print
    elif args.output is None:
        handler = SameFile(args.infile)
    else:
        handler = OtherFile(args.output)

    if args.output_format == "json":
        formatter = _json_dumps
    elif args.output_format == "confuse":
        formatter = Writer().dumps
    elif args.output_format == "table":
        formatter = _format_table

    formatted = formatter(value)
    handler(formatted)


def main(sysargs):
    """Program entry point.

    :param sysargs: The raw command-line arguments (usually obtained
        from :attr:`sys.argv`)
    :type sysargs: Iterable[str]

    :returns: Exit code with the usual semantics (non-zero exit code =
        error).
    :rtype: int
    """
    init()

    args = create_parser().parse_args(sysargs)
    logging.basicConfig(level=getattr(logging, args.verbosity.upper()))

    try:
        logging.info("Processing %s", sysargs)
        logging.info("Parings %s", args.infile)
        if hasattr(args, "query"):
            logging.info("Processing query %s", args.query)
        else:
            logging.info("Processing pattern %s", args.match)
        if hasattr(args, "value"):
            logging.info("Processing value %s", args.value)
        if args.infile == "-":
            parsed = Parser().parse(
                io.TextIOWrapper(sys.stdin.buffer, encoding="utf-8"),
            )
        else:
            with io.open(args.infile, encoding="utf-8") as i:
                parsed = Parser().parse(i.read())
        in_format = in_formatter(args.input_format)
        if args.action == "select":
            if args.output is None:
                args.output = "-"
            if args.query is not None:
                value = parsed.select(*args.query)
            else:
                value = parsed.match(MultiPattern(args.match))
            output(value, args)
        elif args.action == "update":
            value = in_format(args.value)
            updated = parsed.update(args.query, value)
            output(updated, args)
        elif args.action == "append":
            value = in_format(args.value)
            updated = parsed.append(args.query, value)
            output(updated, args)
        elif args.action == "remove":
            if args.output is None:
                args.output = "-"
            if args.query is not None:
                updated = parsed.remove(*args.query)
            else:
                updated = parsed.remove_matching(MultiPattern(args.match))
            output(updated, args)
        return 0
    except Exception as e:
        if logging.root.isEnabledFor(logging.INFO):
            logging.exception(e)
        else:
            logging.error(repr(e))
        return 1
