# -*- coding: utf-8 -*-

from io import StringIO

from lark.lexer import Token

from .compat import str_type
from .formatters import TokensList


class Blank(object):
    """Represents blank space (including newlines).  This is an
    "artifical" token, not really generated by the parser.  Whitespace
    in general, and in particular in the case of LALR parsers presents
    a problem: if the grammar includes whitespace (rather than ignores
    it), then from the perspective of human reader, such grammar
    becomes very cluttered and the parser loses some of its predictive
    power.  So, traditionally, whitespace is ignored when defining
    grammars for such parsers.  We, however, need to keep this
    information so that we don't destroy the original formatting of
    the document once it was parsed.

    The blanks are generated after the initial parsing step where the
    parse tree is extracted.  The original document is then iterated
    over, token after token, and the stretches of text unaccounted for
    are then saved as blanks.  This is true for both whitespace and
    comments.
    """

    def __init__(self, contents, pos):
        self.contents = contents
        self.pos = pos

    @property
    def start_pos(self):
        """
        Imitates similar property from :attr:`~lark.Token.start_pos`
        """
        return self.pos

    @property
    def end_pos(self):
        """
        Imitates similar property from :attr:`~lark.Token.end_pos`
        """
        return self.pos + len(self.contents)

    def __str__(self):
        return self.contents


class Comment(Blank):
    """A special case of :class:`~.Blank` to represent comments."""
    pass


class Tokens(object):
    """Base class for various structural elements found in confuselib
    grammar.

    This class stores original tokens as captured by the parser
    together with the entire body of the document to facilitate
    extraction of whitespace elements.

    :ivar tokens: The tokens (and other elements) that constitute this
        entity.
    :type tokens: Iterable[Union[~.Tokens, ~lark.Token]]

    :ivar body: The entire body of the parsed document
    :type body: str
    """

    def __init__(self, tokens, body):
        self.tokens = tokens
        self.body = body

    def walk(self, dont_expand=()):
        """Iterates over all tokens recursively and yields them to the
        caller.

        :param dont_expand: Additionally, it's possible to request not
            to expand certain elements that answer to
            `isinstance(element, dont_expand)`
        :type dont_expand: Union[type, Tuple[type, ...]]

        :returns: Nothing.  This is an iterator.
        :rtype: Iterable[Union[~.Tokens, ~lark.Token]]
        """
        end = None
        for t in self.tokens_rec(self.tokens):
            if (end is not None) and (end < t.start_pos):
                yield Blank(self.body[end:t.start_pos], end)
            if dont_expand and isinstance(t, dont_expand):
                yield t
            elif hasattr(t, "walk"):
                for tt in t.walk(dont_expand):
                    yield tt
            elif t.isspace():
                yield Blank(t, t.start_pos)
            else:
                yield t
            end = t.end_pos

    def tokens_rec(self, tokens):
        """Helper method to implement recursive extraction of tokens from the
        (possibly nested) parse tree.

        :param tokens: Either a single token or a tree of tokens.
        :type tokens: Union[~.Tokens, ~lark.Token, ~lark.Tree]

        :returns: Nothing.  This is an iterator.
        :rtype: Iterable[~lark.Token]
        """
        if isinstance(tokens, (str_type(), Tokens)):
            yield tokens
        elif hasattr(tokens, "children"):
            for t in self.tokens_rec(tokens.children):
                for t in self.tokens_rec(t):
                    yield t
        else:
            for t in tokens:
                for tt in self.tokens_rec(t):
                    yield tt

    @property
    def start_pos(self):
        """
        Imitates similar property from :attr:`~lark.Token.start_pos`
        """
        return tuple(self.tokens_rec(self.tokens))[0].start_pos

    @property
    def end_pos(self):
        """
        Imitates similar property from :attr:`~lark.Token.end_pos`
        """
        return tuple(self.tokens_rec(self.tokens))[-1].end_pos

    def __str__(self):
        elts = self.tokens_rec(self.tokens)
        buf = StringIO()

        end = None
        for elt in elts:
            if end is not None:
                # TODO(olegs): We need to do this once right after we
                # parse so that we remember all the comments we
                # ignored so far.
                buf.write(self.body[end:elt.start_pos])
            end = elt.end_pos
            buf.write(str_type()(elt))

        buf.seek(0)
        return buf.read()

    def __len__(self):
        return len(str(self))


class Assignment(Tokens):
    """Represents grammar elements of the form `lhs = rhs` (that is
    assignments not including blocks).
    """

    def key(self):
        """The lhs of this grammar element.

        :rtype: ~lark.Token
        """
        return tuple(self.tokens_rec(self.tokens))[0]

    def value(self):
        """The rhs of this grammar element.

        :rtype: Union[~lark.Token, ~.Tokens]
        """
        return tuple(self.tokens_rec(self.tokens))[-1]

    def is_append(self):
        """True if this assignment actually extends the list (uses +=)

        :retype: bool
        """
        for t in self.tokens_rec(self.tokens):
            if t.type == "AOP":
                return t.value == "+="

    def update_value(self, new, formatter):
        """Creates new :class:`~.Assignments` with the new rhs value.

        :param new: The new rhs value.
        :type new: Union[dict, list, str, int, float, bool]

        :param formatter: Functions used to generate literal
            representation of the value.
        :type formatter: Callable[
            Union[dict, list, str, int, float, bool],
            int,
            List[~lark.Token]]

        :returns: A new assignment element with the same lhs as this
                  one, but the new value.
        :rtype: ~cm_confuse.elements.Assignment
        """
        v = self.tokens[-1]
        updated = v.update(new, formatter)
        return Assignment(self.tokens[:-1] + (updated,), updated.body)

    @staticmethod
    def new(key, value, formatter, pos):
        """Creates a new instance of `~.Assignment` with the given key and
        value.

        :param key: The new key to use.
        :type key: str

        :param value: The new value to use.
        :type value: Union[dict, list, str, int, float, bool]

        :param formatter: Functions used to generate literal
            representation of the value.
        :type formatter: Callable[
            Union[dict, list, str, int, float, bool],
            int,
            List[~lark.Token]]

        :param pos: The starting position that will be used to
            generate tokens for this assignment.
        :type pos: int

        :retunrs: New assignment.
        :rtype: ~cm_confuse.elements.Assignment
        """
        tokens = TokensList(pos)
        tokens.add(u"NQ_STRING", key)
        tokens.add(u"WS_INLINE", u" ")
        tokens.add(u"EQL", u"=")
        tokens.add(u"WS_INLINE", u" ")
        for t in formatter(value, tokens.pos):
            tokens.add(t.type, t.value)
        tokens.add(u"NEWLINE", u"\n")
        return Assignment(tokens.tokens, u"")


class Function(Tokens):
    """Represents grammar element that describes a function call."""

    def key(self):
        """Returns the name of the function being called.

        :returns: The name of the function.
        :rtype: ~lark.Token
        """
        return next(iter(self.tokens_rec(self.tokens)))

    def args(self):
        """Returns the arguments of the function.

        :returns: Tuple containing the arguments with which the
                  function was called.
        :rtype: Tuple[~cm_confuse.elements.Value]
        """
        return tuple(
            v.value() for v in self.tokens_rec(self.tokens)
            if type(v) is Value
        )


class Block(Tokens):
    """Represents grammar element that describes a structure with zero or
    more nested assignments or other structures.
    """

    def key(self):
        """If this block is at the toplevel, this is the first name on the
        lhs.  I think the official documentation calls this "section
        name".  If this is inside another block, this is what would've
        been the second name at the toplevel.

        :returns: The leftmost name for the toplevel blocks.
        :rtype: ~lark.Token

        """
        return next(iter(self.tokens_rec(self.tokens)))

    def name(self):
        """This is mostly relevant at the toplevel, this is the second name
        before the brace.

        :returns: The second name before the brace.
        :rtype: ~lark.Token
        """
        seen_key = False
        for t in self.tokens_rec(self.tokens):
            if str(t) == "{":
                break
            if (type(t) is Token) and (t.type == "NQ_STRING"):
                if seen_key:
                    return t
                seen_key = True

    def elements(self):
        """Returns a tuple with all the nested definitions of this block.

        :retunrs: Tuple with nested defintions.
        :rtype: Tuple[~cm_confuse.elements.Assignment]
        """
        # TODO(olegs): Are internal blocks also captured by
        # assignments?
        return tuple(
            v for v in self.tokens_rec(self.tokens)
            if type(v) is Assignment
        )


class Value(Tokens):
    """Either atomic type (what would be a string in libconfuse) or a
    list.

    In this parser implementation we decided to immediately parse
    values into some predefined types.  This means certain differences
    from the way libconfuse interprets the values.  From the
    prespective of libconfuse, there are only strings, lists and
    blocks (structs).  It also provides some pre-defined functions and
    macros for recognizing values s.a. integers, booleans etc.
    However, from the perspective of libconfuse both `"123"` and `123`
    represent the same value: if you request it to be an integer, it's
    going to be an integer, if you request it to be a boolean... well,
    probably it will be a boolean.  Or, maybe you'll get an error.

    Anyways, in practice, we'd probably want to distinguish between
    `"123"` and `123`, and since we don't know how the other program
    reading this configuration is going to interpret the values
    written in it, this seems like a useful heuristic.

    Worth noting that, unlike in Python, there are plenty of ways to
    represent booleans in libconfuse.  Some of the examples include:
    ``Yes``, ``"YES"``, ``'\\y\\E\\s'``, ``"\\131\\105\\123"``,
    similarly for ``"tRUe"`` and ``oN``.  Truth be told, our parser
    doesn't catch all of them.  Again, the assumption was that people
    will not intentionally create "strange" configurations.

    :ivar v: The Python value that was inferred from the parse.
    :type v: Union[list, int, float, str, bool]
    """

    def __init__(self, v, tokens, body):
        super(Value, self).__init__(tokens, body)
        self.v = v

    def value(self):
        """Returns the Python analogue of the libconfuse value (with the
        assumptions mentioned above about hwo that value may be
        represented).

        :returns: The value assigned at parse time.
        :rtype: Union[list, int, float, str, bool]
        """
        return self.v

    def update(self, new, formatter):
        """Clones this value taking into account the old position (in the
        original document), while replacing the old value with the new one.

        :param new: The new value to use.
        :type new: Union[list, int, float, str, bool]

        :returns: New value wrapping the given value at the same
                  position as the old one.
        :rtype: ~cm_confuse.elements.Value
        """
        prefix = self.body[:self.start_pos]
        suffix = self.body[self.end_pos:]
        new_raw = formatter(new, self.start_pos)
        new_body = prefix + "".join(new_raw) + suffix
        return Value(new, new_raw, new_body)


class Item(Tokens):
    """Utility class that represents a toplevel definition (either
    assignment, block or function)
    """

    def item(self):
        """
        Returns the underlying toplevel definition.

        :returns: The underlying toplevel definition.
        :rtype: Union[
            ~cm_confuse.elements.Assignment,
            ~cm_confuse.elements.Block,
            ~cm_confuse.elements.Function]
        """
        for t in self.tokens_rec(self.tokens):
            if isinstance(t, Tokens):
                return t
