# -*- coding: utf-8 -*-

from .elements import Assignment, Blank, Block, Comment, Function, Item, Value


def blank(token):
    """Filter for extracting whitespace elements.

    :param token: The token to be assessed by the filter.
    :type token: Union[~lark.Token, ~cm_confuse.elements.Tokens]

    :returns: Nothing, this is an iterator.
    :rtype: Iterable[~cm_confuse.elements.Blank]
    """
    if isinstance(token, Blank):
        yield token


def identity(token):
    """Identity filter.  This will simply yield the elements given to it.

    :param token: The token to be assessed by the filter.
    :type token: Union[~lark.Token, ~cm_confuse.elements.Tokens]

    :returns: Nothing, this is an iterator.
    :rtype: Iterable[Union[~lark.Token, ~cm_confuse.elements.Tokens]]
    """
    yield token


class CommentFilter(object):
    """Filter for extracting comments.  Since our parser is not set to
    capture comments by itself.  We need to do a second pass, if we
    want to extract the comments from the document.

    :ivar leftover: Some comments may be spread over multiple tokens
        (in particular, the newline is a separate token, but in order
        to parse the comments we need to make sure we saw a newline).
        This field stores the comment prefix (if any) remaining since
        the previous comment.
    :type leftover: str

    :ivar head: The position in the body (so, not relative to the
        token) where the comment starts.
    :type head: int

    :ivar tail: The sequence of characters that are expected to end
        the comment.
    :type tail: str
    """

    comments = {
        "#": "\n",
        "//": "\n",
        "/*": "*/",
    }
    """
    Possible beginning and matching end combinations of comments.
    """

    def __init__(self):
        self.leftover = ""
        self.head = None
        self.tail = None

    def _reset(self):
        self.leftover = ""
        self.head = None
        self.tail = None

    def _matched(self, i, raw):
        if self.head is None:
            for k, v in self.comments.items():
                end = i + len(k)
                if k == raw[i:end]:
                    self.head = i
                    self.tail = v
                    return False, end
            return False, i + 1
        end = i + len(self.tail)
        if raw[i:end] == self.tail:
            return True, end
        return False, i + 1

    def _make_comment(self, i, raw, token):
        return Comment(
            self.leftover + raw[self.head:i],
            token.start_pos + self.head - len(self.leftover),
        )

    def __call__(self, token):
        if isinstance(token, Blank):
            raw = str(token)
            i = 0
            while True:
                found, i = self._matched(i, raw)
                if found:
                    yield self._make_comment(i, raw, token)
                    self._reset()
                if i >= len(raw):
                    break
        if self.head is not None:
            self.leftover += raw[self.head:]


# TODO(olegs): Eventually, insert filter will have to extend this too.
class TransformFilter(object):
    """Base class for various filters that modify the original document by
    either appending more information to it, removing elements or
    inserting new elements.

    :ivar query: The path to the element being modified.
    :type query: Iterable with one or more strings followed by zero
        or more integers.

    :ivar value: New value ot insert or replace (if any).
    :type value: Union[dict, list, float, int, str, bool]

    :ivar formatter: A function to render the value into the new document.
    :type formatter: Callable[
        Union[dict, list, str, int, float, bool],
        int,
        List[~lark.Token]]
    """

    def __init__(self, query, value, formatter):
        self.query = query
        self.value = value
        self.formatter = formatter

    def __call__(self, token):
        if isinstance(token, Blank):
            yield token
        elif isinstance(token, Item):
            for t in self(token.item()):
                yield t
        else:
            for t in self.transform(token):
                yield t

    def _iter_values(self, token, n):
        i = 0
        for t in token.walk(Value):
            found = False
            if type(t) is Value:
                found = (i == n)
                i += 1
            yield found, t
        if n > i - 1:
            raise IndexError(
                "{} is out of bounds for {}".format(n, token.value()),
            )

    def match_assignment(self, token):
        """Entry point for the child classes.  This is where the
        decision is made whether the value that needs work is found,
        and what to do about it: replace entirely, replace sub-element
        at index, or leave it as is.

        :param token: This is either :class:`~lark.Token` or something
            that looks like it.  It's important that it has
            `start_pos` and `end_pos` properties.
        :type token: Union[
            ~lark.Token,
            ~cm_confuse.elements.Tokens,
            ~cm_confuse.elements.Blank]

        :returns: Nothing. This is a generator.
        :rtype: Iterable[
            ~lark.Token,
            ~cm_confuse.elements.Tokens,
            ~cm_confuse.elements.Blank]
        """
        # TODO(olegs): This doesn't deal properly with incremental
        # assignment nor with duplicates.
        if token.key() == self.query[0]:
            if len(self.query) == 1:
                for t in self.update_value(token):
                    yield t
            else:
                for t in self.update_index(token):
                    yield t
        else:
            yield token


class UpdateFilter(TransformFilter):
    """Filter used to replace existing values in the document with the
    new ones.
    """

    def update_value(self, token):
        """Superclass expects children to implement this.  This is
        called when the value needs to be updated.

        :param token: This is either :class:`~lark.Token` or something
            that looks like it.  It's important that it has
            `start_pos` and `end_pos` properties.
        :type token: Union[
            ~lark.Token,
            ~cm_confuse.elements.Tokens,
            ~cm_confuse.elements.Blank]

        :returns: Nothing. This is a generator.
        :rtype: Iterable[
            ~lark.Token,
            ~cm_confuse.elements.Tokens,
            ~cm_confuse.elements.Blank]
        """
        yield token.update_value(self.value, self.formatter)

    def _update_index_rec(self, token, pos, indices):
        for found, t in self._iter_values(token, pos):
            if found and indices:
                for t in self._update_index_rec(t, indices[0], indices[1:]):
                    yield t
            elif found:
                yield t.update(self.value, self.formatter)
            else:
                yield t

    def update_index(self, token):
        """Superclass expects children to implement this.  This is
        called when an element at index inside value needs to be updated.

        :param token: This is either :class:`~lark.Token` or something
            that looks like it.  It's important that it has
            `start_pos` and `end_pos` properties.
        :type token: Union[
            ~lark.Token,
            ~cm_confuse.elements.Tokens,
            ~cm_confuse.elements.Blank]

        :returns: Nothing. This is a generator.
        :rtype: Iterable[
            ~lark.Token,
            ~cm_confuse.elements.Tokens,
            ~cm_confuse.elements.Blank]
        """
        # First time we iterate through the Assignment rather than the
        # list, that's why we use 0 as an artificial index, since
        # assignment will have only one element of type Value, and
        # that is the list we want to modify.
        for t in self._update_index_rec(token, 0, self.query[1:]):
            yield t

    def transform(self, token):
        """This is the entry point where superclass has done its job
        filtering irrelevant elements and transferred control to the
        sublcass.

        :returns: Nothing, this is an iterator.
        :rtype: Iterable[Union[~lark.Token, ~cm_confuse.elements.Tokens]]
        """
        if isinstance(token, Assignment):
            for t in self.match_assignment(token):
                yield t
        elif isinstance(token, Block):
            # TODO(olegs): This needs to be tested, plus see the TODO
            # in match_assignment()
            if token.key() == self.query[0]:
                if len(self.query) == 1:
                    yield token.update_value(self.value, self.formatter)
                else:
                    self.query = self.query[1:]
                    for tk in token.tokens:
                        for t in self(tk):
                            yield t
            else:
                yield token
        elif isinstance(token, Function):
            # TODO(olegs): Not sure what to do about functions in this
            # context.
            yield token


class RemoveFilter(TransformFilter):

    def update_value(self, token):
        """Superclass expects children to implement this.  This is
        called when the value needs to be removed.

        :param token: This is either :class:`~lark.Token` or something
            that looks like it.  It's important that it has
            `start_pos` and `end_pos` properties.
        :type token: Union[
            ~lark.Token,
            ~cm_confuse.elements.Tokens,
            ~cm_confuse.elements.Blank]

        :returns: Nothing. This is a generator.
        :rtype: Iterable[None]
        """
        yield None

    def _update_index_rec(self, token, pos, indices):
        for found, t in self._iter_values(token, pos):
            if found and indices:
                for tt in self._update_index_rec(t, indices[0], indices[1:]):
                    yield tt
            elif found:
                yield None
            else:
                yield t

    def update_index(self, token):
        """Superclass expects children to implement this.  This is
        called when an element at index inside value needs to be updated.

        :param token: This is either :class:`~lark.Token` or something
            that looks like it.  It's important that it has
            `start_pos` and `end_pos` properties.
        :type token: Union[
            ~lark.Token,
            ~cm_confuse.elements.Tokens,
            ~cm_confuse.elements.Blank]

        :returns: Nothing. This is a generator.
        :rtype: Iterable[
            ~lark.Token,
            ~cm_confuse.elements.Tokens,
            ~cm_confuse.elements.Blank,
            None]
        """
        for t in self._update_index_rec(token, 0, self.query[1:]):
            yield t

    def transform(self, token):
        """This is the entry point where superclass has done its job
        filtering irrelevant elements and transferred control to the
        sublcass.

        :returns: Nothing, this is an iterator.
        :rtype: Iterable[Union[~lark.Token, ~cm_confuse.elements.Tokens, None]]
        """
        if isinstance(token, Assignment):
            for t in self.match_assignment(token):
                yield t
        elif isinstance(token, Block):
            # TODO(olegs): This needs to be tested, plus see the TODO
            # in match_assignment()
            if token.key() == self.query[0]:
                if len(self.query) == 1:
                    yield None
                else:
                    self.query = self.query[1:]
                    for tk in token.tokens:
                        for t in self(tk):
                            yield t
            else:
                yield token
        elif isinstance(token, Function):
            # TODO(olegs): Not sure what to do about functions in this
            # context.
            yield token


class KVFilter(object):

    # TODO(olegs): Function may be treated as a compound, but let's
    # not do this for now.
    terminals = Blank, Comment, Function

    def __init__(self, separator="/"):
        self.separator = separator

    def path(self, p):
        return self.separator.join(str(elt) for elt in p)

    def clone(self):
        _, cargs, state = self.__reduce__()
        cls = cargs[0]
        result = cls(**state)

        return result

    def __call__(self, path, token):
        p = self.path(path)
        if isinstance(token, self.terminals):
            yield p, token
        elif isinstance(token, Assignment):
            for t in self.clone()(path + (token.key(),), token.value()):
                yield t
        elif isinstance(token, Item):
            for t in self.clone()(path, token.item()):
                yield t
        elif isinstance(token, Block):
            # TODO(olegs): deal with toplevel blocks
            for t in self.clone()(path + (token.key(),), token.elements()):
                yield t
        else:
            yield p, token
