# -*- coding: utf-8 -*-

from lark.lexer import Token

from .compat import str_type


def default_bool_formatter(val, pos):
    """Given Python value that's assumed to be boolean, this will create a
    tuple with a single element that's a token representing boolean
    value.  This will always result in literal string `true` or `false`.

    :param val: The Python value that should be represented.
    :type val: bool

    :param pos: The position at which the value will be inserted into
        the document.
    :type pos: int

    :returns: Tuple with a single element being the token representing
        the boolean value.
    :rtype: Tuple[~lark.Token]
    """
    raw = str(val).lower()
    result = Token("BOOLEAN", raw)
    result.start_pos = pos
    result.end_pos = pos + len(raw)
    return result,


class ValueFormatter(object):
    """Base class for various value formatters.

    :ivar string_formatter: It's easier to use `repr()` as a means to
        format strings, however, our version of configuation typically
        uses double quotes, this calls for different implementations
        of value formatters.
    :type string_formatter: Callable[str, int, Tuple[~lark.Token]]
    """

    def __init__(self, string_formatter=None):
        if string_formatter is None:
            self.string_formatter = self.default_string_formatter
        else:
            self.string_formatter = string_formatter

    def __call__(self, val, pos):
        ttype = None
        if isinstance(val, (int, float)):
            ttype = "NUMBER"
            raw = repr(val)
        elif isinstance(val, str_type()):
            ttype = "STRING"
            raw = self.string_formatter(val, 0)[0]
        else:
            raise NotImplementedError(
                "Unsupported value type: {}, "
                "you need to implement custom formatter for {}".format(
                    type(val),
                    val,
                ))
        result = Token(ttype, raw)
        result.start_pos = pos
        result.end_pos = pos + len(raw)
        return result,

    def default_string_formatter(self, val, pos):
        result = repr(val)
        if result.startswith("u"):
            # Python 2 way of printing unicode
            result = result[1:]
        return str_type()(result),


def dq_string_formatter(val, pos):
    """Formats a Python string to have double quotes.  It's somewhat
    defective because it will not unquote single quote, but the parser
    should be OK with it.

    :param pos: The position at which the value will be inserted into
        the document.
    :type pos: int

    :returns: Tuple with a single element being the token representing
        the string value.
    :rtype: Tuple[~lark.Token]
    """
    pieces = ["\""]
    for c in repr(val)[1:-1]:
        if c == "\"":
            pieces.append("\\\"")
        else:
            pieces.append(c)
    pieces.append("\"")
    raw = "".join(pieces)
    result = Token("STRING", raw)
    result.start_pos = pos
    result.end_pos = pos + len(raw)
    return result,


default_value_formatter = ValueFormatter()
"""Formats a Python value, s.a. string, integer, floating point
number.  Strings will use single quote sintax.

:param pos: The position at which the value will be inserted into
    the document.
:type pos: int

:returns: Tuple with a single element being the token representing
    the string value.
:rtype: Tuple[~lark.Token]
"""


cmd_value_formatter = ValueFormatter(dq_string_formatter)
"""Formats a Python value, s.a. string, integer, floating point
number.  Strings will use single quote sintax.

:param pos: The position at which the value will be inserted into
    the document.
:type pos: int

:returns: Tuple with a single element being the token representing
    the string value.
:rtype: Tuple[~lark.Token]
"""


class TokensList(object):
    """Utility class to make creating lists of tokens easier."""

    def __init__(self, pos):
        self.tokens = []
        self.tail = pos

    def add(self, ttype, value):
        token = Token(ttype, value)
        self.tokens.append(token)
        token.start_pos = self.tail
        token.end_pos = self.tail + len(token)
        self.pos = token.end_pos


class ListFormatter(object):
    """Base class for creating list formatters.

    Similar to :class:`~.ValueFormatter`, items in the list, provided
    they are strings may use different quotations marks.  This class
    allows creation of formatters with customized string formatter.
    Similarly, it allows customization of boolean formatter (eg. if
    you want `YES` or `NO` rather than `true` and `false`).

    :ivar bool_formatter: Formatter to use to format boolean values.
    :type bool_formatter: Callable[bool, int, Tuple[~lark.Token]]

    :ivar val_formatter: Formatter to use to format other values.
    :type val_formatter: Callable[
        Union[str, int, float],
        int,
        Tuple[~lark.Token]]
    """

    def __init__(
            self,
            bool_formatter=default_bool_formatter,
            val_formatter=default_value_formatter,
    ):
        self.bool_formatter = bool_formatter
        self.val_formatter = val_formatter

    def __call__(self, val, pos):
        tokens = TokensList(pos)
        tokens.add("LBRACE", "{")
        tokens.add("SPACE", " ")
        comma = False

        # TODO(olegs): Format nested lists
        for v in val:
            if comma:
                tokens.add("COMMA", ",")
                tokens.add("SPACE", " ")
            if isinstance(v, bool):
                tv = self.bool_formatter(v, 0)[0]
                tokens.add(tv.type, tv.value)
            elif isinstance(v, (int, float, str_type())):
                tv = self.val_formatter(v, 0)[0]
                tokens.add(tv.type, tv.value)
            else:
                raise TypeError("Unsupported list element type: {type(v)}")
            comma = True

        tokens.add("SPACE", " ")
        tokens.add("RBRACE", "}")
        return tokens.tokens


default_list_formatter = ListFormatter()
"""Formats a Python list.  Strings inside this list will use single
quote sintax.

:param pos: The position at which the value will be inserted into
    the document.
:type pos: int

:returns: Tuple with a single element being the token representing
    the string value.
:rtype: Tuple[~lark.Token]
"""

cmd_list_formatter = ListFormatter(default_bool_formatter, dq_string_formatter)
"""Formats a Python list.  Strings inside this list will use double
quote sintax.

:param pos: The position at which the value will be inserted into
    the document.
:type pos: int

:returns: Tuple with a single element being the token representing
    the string value.
:rtype: Tuple[~lark.Token]
"""


# TODO(olegs): This hasn't been tested, this will also need versions
# with verid booleans and different kidns of strings.
def default_dict_formatter(val, pos):
    """
    Formats a Python dict.  Strings will use single quote sintax.

    :param pos: The position at which the value will be inserted into
        the document.
    :type pos: int

    :returns: Tuple with a single element being the token representing
        the string value.
    :rtype: Tuple[~lark.Token]
    """
    tokens = TokensList(pos)
    tokens.add("LBRACE", "{")
    tokens.add("NEWLINE", "\n")
    endl = False

    for k, v in val.items():
        if endl:
            tokens.add("NEWLINE", "\n")

        if not isinstance(k, str_type()):
            raise TypeError("Unsupported dict key type: {type(k)}")
        tokens.add("NQ_STRING", k)
        tokens.add("SPACE", " ")

        # TODO(olegs): There's some duplication here that can be
        # removed.
        if isinstance(v, bool):
            tokens.add("EQL", "=")
            tokens.add("SPACE", " ")
            tv = default_bool_formatter(v, 0)
            tokens.add(tv.type, tv.value)
        elif isinstance(v, (int, float, str_type())):
            tokens.add("EQL", "=")
            tokens.add("SPACE", " ")
            tv = default_value_formatter(v, 0)
            tokens.add(tv.type, tv.value)
        elif isinstance(v, list):
            tokens.add("EQL", "=")
            tokens.add("SPACE", " ")
            for t in default_list_formatter(v, 0):
                tokens.add(t.type, t.value)
        elif isinstance(v, dict):
            for t in default_dict_formatter(v, 0):
                tokens.add(t.type, t.value)
        else:
            raise TypeError("Unsupported dict element type: {type(v)}")
        endl = True

    tokens.add("RBRACE", "}")
    return tokens.tokens
