# -*- coding: utf-8 -*-

import ast
import copy
import json
import logging
import re
from collections import defaultdict
from io import StringIO

from lark import Lark, Transformer, v_args
from lark.tree import Tree

from .compat import str_type
from .elements import Assignment, Blank, Block, Function, Item, Value
from .filters import (
    CommentFilter,
    KVFilter,
    RemoveFilter,
    UpdateFilter,
    blank,
    identity
)
from .formatters import (
    cmd_list_formatter,
    cmd_value_formatter,
    default_bool_formatter,
    default_dict_formatter,
    default_list_formatter,
    default_value_formatter,
    dq_string_formatter
)


confuse_grammar = r"""
    EQL: "="
    APPEND: "+="
    LBRACE: "{"
    RBRACE: "}"
    LPAREN: "("
    RPAREN: ")"
    COMMA: ","
    AOP: EQL | APPEND
    SQ_STRING: /'([^'\\]*(\\'|\\\\|\\\n)*)*'/
    NQ_STRING: /([^ "'#\/(){}=\t\r\n+,\$]|\/[^ "'#\/(){}=\t\r\n+,*\$])+/
    ENV_VAR: "${" /[^}]+/ "}"
    BOOLEAN: /true|false|on|off|no|yes/i

    ?start: items -> result
    ?items: item*
    ?item: assignment    -> item
         | block         -> item
         | function      -> item
         | NEWLINE

    ?assignment: key AOP value -> assignment

    ?block: NQ_STRING block_value -> block

    ?function: NQ_STRING LPAREN values NEWLINE* RPAREN -> function
             | NQ_STRING LPAREN NEWLINE* RPAREN        -> function

    ?key: NQ_STRING

    ?block_value: NQ_STRING LBRACE assignments? RBRACE
    ?assignment_item: assignment
                    | block_value
    ?assignments: assignment_item? ( NEWLINE assignment_item? )*

    ?values: valuenl ( commanl valuenl )* commanl?
    ?valuenl: value NEWLINE*
    ?commanl: COMMA NEWLINE*

    ?value: number
          | string
          | env_var
          | boolean
          | list

    ?boolean: BOOLEAN -> boolean

    ?env_var: ENV_VAR -> env_var

    ?number: NUMBER -> number

    ?string: STRING    -> string
           | SQ_STRING -> string
           | NQ_STRING -> string

    ?list: list_start list_end -> list
    ?list_start: LBRACE NEWLINE*
    ?list_end: values NEWLINE* RBRACE
             | NEWLINE* RBRACE

    %import common.ESCAPED_STRING -> STRING
    %import common.SIGNED_NUMBER -> NUMBER
    %import common.CNAME -> NAME
    %import common.WS_INLINE
    %import common.NEWLINE
    %import common.SH_COMMENT
    %import common.CPP_COMMENT
    %import common.C_COMMENT
    %ignore WS_INLINE
    %ignore SH_COMMENT
    %ignore CPP_COMMENT
    %ignore C_COMMENT
"""
"""
Definition of libconfuse grammar

:meta hide-value:
"""


class ParseResult(object):
    """Represents result of parsing configuration that uses libconfuse
    format.

    It doesn't normally make sense to create instances of this class
    in user code.  This is normally a result of calling
    :meth:`~.Parser.parse()` or some of the parser's method that
    modify the result.

    There are a number of difficulties resulting from the need of
    keeping information about the parsed document, s.a. comments and
    the whitespace that make it difficult to modify the parser result
    in place.  Therefore methods that would usually modify objects in
    place (s.a. update or delete) instead produce new
    :class:`~.ParseResult` objects with necessary modifications.  It's
    probably a bad idea to modify the members of this object.

    :ivar definitions: The collection of the elements representing the
        parse. At the top level in this collection, you'll find
        mostely :class:`~cm_confuse.elements.Item` objects, as well as
        some whitespace objects.  Items may contain
        :class:`~cm_confuse.elements.Assignment` objects or
        :class:`~cm_confuse.elements.Block` objects and so on.  See
        :const:`~cm_confuse.parser.confuse_grammar` for precise
        description of possible hierarchies.
    :type definitions: :class:`cm_confuse.elements.Tokens`

    :ivar body: The raw contents of the parsed file.
    :type body: str
    """

    def __init__(self, definitions, body):
        if definitions:
            if isinstance(definitions[0], Tree):
                self.definitions = definitions[0].children
            elif isinstance(definitions[0], Item):
                self.definitions = definitions
            else:
                raise ValueError(
                    "Unexpected parse: {}".format(type(definitions[0])),
                )
        else:
            self.definitions = ()
        self.body = body

        self._data = {}
        self._functions = defaultdict(list)

        self._parse_blocks()

    def data(self):
        return copy.deepcopy(self._data)

    def _walk(self, transformer):
        if self.definitions:
            end = self.definitions[0].start_pos
            for t in transformer(Blank(self.body[0:end], 0)):
                yield t
            last = None
            for d in self.definitions:
                if last is not None:
                    for t in transformer(
                            Blank(self.body[last:d.start_pos], last),
                    ):
                        yield t
                if not hasattr(d, "walk"):
                    if d.isspace():
                        for t in transformer(Blank(d, d.start_pos)):
                            yield t
                    else:
                        for t in transformer(d):
                            yield t
                    last = d.end_pos
                else:
                    for b in d.walk():
                        for t in transformer(b):
                            yield t
                    last = d.end_pos
            if last is not None:
                for t in transformer(
                        Blank(self.body[last:len(self.body)], last),
                ):
                    yield t
        else:
            for t in transformer(Blank(self.body, 0)):
                yield t

    def walk(self, transformer):
        """Iterate over all the tokens of the parse applying `transformer` to
        them in such a way that if `transformer` yields `None` the
        result is skipped.  Non-empty results are yielded back.

        .. code-block:: python

            result = parser.parse(...)
            for comment in result.walk(CommentFilter()):
                print("This is a comment")
                print(comment)

        :param transformer: A generator, i.e. a function with `yield`
            in its definition or a class that implements `__call__`
            protocol with `yield` in the implementation.  This
            function will be called with the token (in the order they
            appear in the parsed document).  It is expected to yield
            back a number of tokens (to be used in creating a modified
            document).  If token is `None`, it is skipped.

        :type transformer: Callable[
            :class:`~lark.Token`, Iterable[:class:`~lark.Token`]

        :returns: This is an iterator, so, it doesn't return anything.
        :rtype: Iterable[:class:`~lark.Token`]
        """
        for t in self._walk(transformer):
            if t is not None:
                yield t

    def _walk_structure(self, transformer):
        if self.definitions:
            end = self.definitions[0].start_pos
            for t in transformer(Blank(self.body[0:end], 0)):
                yield t
            last = None
            for d in self.definitions:
                if last is not None:
                    for t in transformer(
                            Blank(self.body[last:d.start_pos], last),
                    ):
                        yield t
                if isinstance(d, str_type()):
                    # This is either a comment or a newline
                    for t in transformer(Blank(d, last)):
                        yield t
                else:
                    for t in transformer(d):
                        yield t
                last = d.end_pos
            if last is not None:
                for t in transformer(
                        Blank(self.body[last:len(self.body)], last),
                ):
                    yield t
        else:
            for t in transformer(Blank(self.body, 0)):
                yield t

    def walk_structure(self, transformer):
        """Similar to :meth:`~.ParseResult.walk()`, but rather than tokens it
        used elements defined in :mod:`cm_confuse.elements` module.

        :param transformer: Same as in :meth:`~.ParseResult.walk()`.
        :type transformer: Callable[
            ~lark.Token, Iterable[~lark.Token]

        :returns: This is a generator, so it doesn't return anything
        :rtype: Iterable[~cm_confuse.elements.Tokens]
        """
        for t in self._walk_structure(transformer):
            if t is not None:
                yield t

    def _walk_path(self, transformer):
        path = "",
        if self.definitions:
            end = self.definitions[0].start_pos
            for t in transformer(path, Blank(self.body[0:end], 0)):
                yield t
            last = None
            for d in self.definitions:
                if last is not None:
                    for t in transformer(
                            path,
                            Blank(self.body[last:d.start_pos], last),
                    ):
                        yield t
                if isinstance(d, str_type()):
                    # This is either a comment or a newline
                    for t in transformer(path, Blank(d, last)):
                        yield t
                else:
                    for t in transformer(path, d):
                        yield t
                last = d.end_pos
            if last is not None:
                for t in transformer(
                        path,
                        Blank(self.body[last:len(self.body)], last),
                ):
                    yield t
        else:
            for t in transformer(path, Blank(self.body, 0)):
                yield t

    def walk_path(self, transformer):
        for t in self._walk_path(transformer):
            if t is not None:
                yield t

    def blanks(self):
        """Iterates over all blank elements in the parsed definition

        :returns: This is a generator, so it doesn't return anything
        :rtype: Iterable[~lark.Token]
        """
        for t in self.walk(blank):
            yield t

    def comments(self):
        """Iterates over all comments in the parsed definition

        :returns: This is a generator, so it doesn't return anything
        :rtype: Iterable[~lark.Token]
        """
        for t in self.walk(CommentFilter()):
            yield t

    def each(self):
        """Iterates over all elements in the parsed definition

        :returns: This is a generator, so it doesn't return anything
        :rtype: Iterable[~lark.Token]
        """
        for t in self.walk(identity):
            yield t

    def _append_or_assign(self, data, assignment):
        key = str(assignment.key())
        if assignment.is_append():
            val = data.get(key)
            if val is None:
                val = []
                data[key] = val
            elif type(val) is not list:
                raise SyntaxError("Cannot append to {}".format(val))
            val.extend(assignment.value().value())
        else:
            data[key] = assignment.value().value()

    def _parse_inner_block(self, elements):
        data = {}
        for d in elements:
            if type(d) is Assignment:
                self._append_or_assign(data, d)
            elif type(d) is Block:
                child = data.get(d.key(), {})
                child[d.name()] = self._parse_inner_block(d.value())
        return data

    def _parse_blocks(self):
        data = self._data
        functions = self._functions
        for d in self.definitions:
            if type(d) is Item:
                item = d.item()
                if type(item) is Assignment:
                    self._append_or_assign(data, item)
                elif type(item) is Function:
                    functions[str(item.key())].append(item.args())
                elif type(item) is Block:
                    # I think functions may only be called at toplevel
                    child = data.get(item.key(), {})
                    child[item.name()] = self._parse_inner_block(
                        item.elements(),
                    )
                    data[str(item.key())] = child

    def _select_rec(self, query, defs):
        if not query:
            return defs

        q, rq = query[0], query[1:]
        try:
            return self._select_rec(rq, defs[q])
        except KeyError:
            if logging.root.isEnabledFor(logging.DEBUG):
                buf = StringIO()
                logging.debug("Dumping configuration")
                json.dump(self._data, buf, indent=2)
                buf.seek(0)
                for line in buf:
                    logging.debug(line.rstrip())
                logging.debug("End dumping configuration")
            raise

    def select(self, *query):
        """Selects a value sepcified by `query`.

        .. code-block:: python

            result = parser.parse(...)
            # Select the second element from the list in slot `bar'
            # of a struct assigned to `foo'
            value = result.select("foo", "bar", 1)

        :param query: Path to the value to extract. It consists of a
            number of lhs names followed by the number of indices into
            the list (if applicable).
        :type query: One more more strings followed by zero or more
            integers.

        :returns: The value specified by `query`.
        :rtype: Union[dict, list, str, float, int, bool]
        """
        return self._select_rec(query, self._data)

    def remove(self, *query):
        # To raise an error if the item cannot be found
        self.select(*query)
        buf = StringIO()
        filter = RemoveFilter(query, None, None)
        for t in self.walk_structure(filter):
            buf.write(str_type()(t))
        buf.seek(0)
        return Parser().parse(buf.read())

    def match(self, pattern, separator="/"):
        if isinstance(pattern, str_type()):
            pattern = re.compile(pattern)
        filter = KVFilter(separator)
        empty = ParseResult(None, "")

        for k, v in self.walk_path(filter):
            if re.search(pattern, k):
                # TODO(olegs): In principle, this can also be a blank
                # or a comment, but we don't yet have a way of
                # building the whole thing back with comments and
                # formatting.
                empty = empty.append(k.split(separator)[1:], v.value())

        return Parser().parse(Writer().dumps(empty))

    def _formatter(self, value, default=None):
        if default is None:
            if isinstance(value, (int, float, str_type())):
                default = default_value_formatter
            elif isinstance(value, bool):
                default = default_bool_formatter
            elif isinstance(value, (list, tuple)):
                default = default_list_formatter
            elif isinstance(value, dict):
                default = default_dict_formatter
            else:
                raise ValueError(
                    "Cannot automatically determine formatter for "
                    "{}({})".format(type(value), value),
                )
        return default

    def update(self, query, value, value_formatter=None):
        """Creates a copy of this object with the value specified by `query`
        replaced with the new `value`.

        :param query: An iterable of one more more strings followed by
            zero or more integers.  Strings are the lhs in assignments
            while integers are indices in lists (if the rhs value is a
            list).
        :type query: One more more strings followed by zero or more integers.

        :param value: The new value to use instead of the old one.
        :type value: Union[dict, list, float, int, str, bool]

        :param value_formatter: A function to use to generate tokens
            necessary to represent the newly inserted value.  This
            function will be called with the value that needs to be
            represented and the position at which it is to be
            inserted.  It is expected to produce a list of tokens.

        :type value_formatter: Callable[
            Union[dict, list, str, int, float, bool],
            int,
            List[~lark.Token]]

        :returns: New :class:`~.ParseResult` with new value replacing
            the old one.
        :rtype: ~cm_confuse.parser.ParseResult
        """
        # To make sure the object can be found
        self.select(*query)
        buf = StringIO()
        filter = UpdateFilter(
            query,
            value,
            self._formatter(value, value_formatter),
        )
        for t in self.walk_structure(filter):
            buf.write(str_type()(t))
        buf.seek(0)
        return Parser().parse(buf.read())

    def append(self, query, value, value_formatter=None):
        """Similar to :meth:`~.ParseResult.update()`, however, appends the
        value rather than replaces the old one.

        :param query: An iterable of one more more strings followed by
            zero or more integers.  Strings are the lhs in assignments
            while integers are indices in lists (if the rhs value is a
            list).
        :type query: One more more strings followed by zero or more integers.

        :param value: The new value to use instead of the old one.
        :type value: Union[dict, list, float, int, str, bool]

        :param value_formatter: A function to use to generate tokens
            necessary to represent the newly inserted value.  This
            function will be called with the value that needs to be
            represented and the position at which it is to be
            inserted.  It is expected to produce a list of tokens.

        :type value_formatter: Callable[
            Union[dict, list, str, int, float, bool],
            int,
            List[~lark.Token]]

        :returns: New :class:`~.ParseResult` with new value replacing
            the old one.
        :rtype: ~cm_confuse.parser.ParseResult
        """
        if len(query) > 1:
            # We are trying to append to one of the existing
            # definitions.  We need to make sure the definition
            # exists, otherwise we won't append anything and will not
            # inform the user nothing was appended.
            self.select(*query)
            # TODO(olegs): Appends to definitions not at the toplevel
            # will not be supported in the context of TF-803.
            return self.insert(query, value, value_formatter)

        return Parser().parse(
            self.body + u"\n" + str_type()(Assignment.new(
                query[0],
                value,
                self._formatter(value, value_formatter),
                len(self.body),
            )))

    def insert(self, how, query, value, value_formatter=None):
        """More general case of :meth:`~.ParseResult.append()`.  Rather than
        adding the value at the end of the file / list / block, this
        will insert it at the specified position.

        :param query: An iterable of one more more strings followed by
            zero or more integers.  Strings are the lhs in assignments
            while integers are indices in lists (if the rhs value is a
            list).
        :type query: One more more strings followed by zero or more integers.

        :param value: The new value to use instead of the old one.
        :type value: Union[dict, list, float, int, str, bool]

        :param value_formatter: A function to use to generate tokens
            necessary to represent the newly inserted value.  This
            function will be called with the value that needs to be
            represented and the position at which it is to be
            inserted.  It is expected to produce a list of tokens.
        :type value_formatter: Callable[
            Union[dict, list, str, int, float, bool],
            int,
            List[~lark.Token]]

        :returns: New :class:`~.ParseResult` with new value replacing the old
            one.
        :rtype: ~cm_confuse.parser.ParseResult
        """
        # TODO(olegs): This is a more general version of append, and,
        # eventually, append will have to use this.
        pass

    def select_functions(self, name, index):
        """Functions (function calls) are stored separately from the rest of
        the values.  It is my understanding that function names are
        allowed to be the same as the lhs of other definitions.
        That's why they cannot be stored in the same dictionary with
        the rest of the data.  It should be also possible to call the
        same function multiple times.  That's why there's `index` parameter.

        :param name: The function name to look for.
        :type name: str

        :param index: The nth call of the function to find.

        :returns: The function call definition.
        :rtype: TODO(olegs): This isn't implemented yet.
        """
        return self._functions[name][index]

    def select_comments(self, index):
        """Finds the `index`'s comment in the document and returns it.

        :param index: The position of the comment to look for.
        :type index: int

        :returns: An element that has the comment text and some
            metadata (s.a. the position) of the comment.
        :rtype: ~cm_confuse.elements.Comment
        """
        for i, c in enumerate(self.walk(CommentFilter())):
            if i == index:
                return str(c)

    def cmd_upsert_list(self, field, key, value):
        """Updates, if found, adds otherwise a new string of the format
        `"{key}={value}"` in the list assigned to `field`.

        For example, if the original document contains:

        .. code-block:: python

            foo = { "bar=1", "baz=2" }

        Given `field = "foo"`, `key == "qux"`, and `value = 4` will
        result in:

        .. code-block:: python

            foo = { "bar=1", "baz=2", "qux=4" }

        However, if `key == "bar"`:

        .. code-block:: python

            foo = { "bar=4", "baz=2" }

        Note that if the list contains multiple values with the same
        key, then the last one is modified.

        :param field: The lhs of toplevel definition (not block) in
            the original document.  Must be assigned a list.
        :type field: str

        :param key: The prefix (possibly the part of the value in the
            list) before equals sign.
        :type key: str

        :param value: The suffix (following the equals sign).  Needs
            to implement both `__repr__` and `__str__`.
        :type value: Either sring or something that may be converted
            to string.

        :returns: New :class:`~ParseResult` with new value replacing
            the old one.
        :rtype: ~cm_confuse.parser.ParseResult
        """
        writer = Writer()
        old = self.select(field)
        # Apparently, the existing configuration doesn't use the
        # feature of the format that allows it to store fields of
        # a struct...  the only observed complex value is a list
        new, update = [], None
        for i, elt in enumerate(old):
            k, v = elt.split("=", 1)
            # The old code believed that elements in this list
            # with the same prefix are not allowed, so it removed
            # them, replacing with the new key-value pair.
            if k.strip() == key:
                # This isn't 100% correct, as strings in
                # libconfuse may be written in many different
                # ways, some of which will prevent this test from
                # succeeding... (eg. escaping would work
                # differently for double-quoted string and
                # singl-quoted strings).  Also, I'm not sure about
                # strip(). I don't know if whitespace is
                # significant here.
                if v == str(value):
                    # The old code didn't restart cmd in this
                    # case, even if restart was requested.
                    return "", False
                # I have no idea what to do if multiple elements
                # have the same prefix.
                update = i
            else:
                new.append(elt)
        # This is the behavior of the old code.  Even if the key
        # or the value have quotes in them, it doesn't try to
        # escape them.
        new_val = "{}={}".format(key, value)
        new.append(new_val)
        if update is None:
            # We didn't find the previous instance of the key.
            #
            # TODO(olegs): Appending, right now only works for
            # toplevel elements, we don't know how to append to an
            # existing nested element.  Once that is implemented,
            # we'll just call append() with appropriate query.
            updated = self.update((field,), new, cmd_list_formatter)
        else:
            updated = self.update((field, i), new_val, dq_string_formatter)
        return writer.dumps(updated), True

    def cmd_upsert(self, key, value):
        """Appends or updates a value in original document.  Right now,
        functions cannot be updated in this way.

        :param key: A toplevel lhs definition.
        :type key: str

        :param value: The value to substitute for the old one.
        :type value: Union[dict, list, float, str, int, bool]
        """
        result = None
        formatter = cmd_value_formatter
        if isinstance(value, list):
            formatter = cmd_list_formatter
        try:
            result = self.update((key,), value, formatter)
        except KeyError:
            result = self.append((key,), value, formatter)
        return Writer().dumps(result)


@v_args(inline=True)
class Parser(Transformer):
    """This class is a transformer object for the Lark parser.  Most
    methods of this class are callbacks that the parser will use to
    process some parse trees as it encounters them.  They are not
    intended to be called from user code.

    Transformer, in this context, means that the object's methods are
    used to modify the parse outcome of the parser.

    The parser is reusable (meaning you can parse different documents
    using the same instance of the parser), but it's not reentrable
    (meaning, you cannot parse multiple documents at once).

    :ivar body: The raw contents of the file to be parsed.
    :type body: str
    """

    def __init__(self):
        self.lark = Lark(
            confuse_grammar,
            parser="lalr",
            transformer=self,
        )
        self.body = u""

    def parse(self, body):
        """Interpret the contens of the body as having libconfuse style
        configuation.

        :param body: The text to parse.
        :type body: str

        :returns: New :class:`~.ParseResult`.
        :rtype: ~cm_confuse.parser.ParseResult
        """
        self.body = body
        return self.lark.parse(body)

    def result(self, *items):
        """
        :meta private:
        """
        return ParseResult(items, self.body)

    def number(self, n):
        """
        :meta private:
        """
        if n.startswith("0"):
            return Value(int(n, 8), n, self.body)
        return Value(ast.literal_eval(n), n, self.body)

    def string(self, s):
        """
        :meta private:
        """
        bool_vals = u"true", u"false", u"on", u"off", u"yes", u"no"
        if s.type == u"NQ_STRING":
            if s.value.lower() in bool_vals:
                # TODO(olegs): We keep BOOLEAN token in the parser
                # even though it is unreachable because of the
                # conflict with NQ_STRING just to have BOOLEAN type.
                # We need to figure out the priority settings in lark
                # to avoid this, and have BOOLEAN terminal be higher
                # priority than NQ_STRING.
                s.type = u"BOOLEAN"
                return self.boolean(s)
            v = u"u'" + s + u"'"
        elif s.type == u"SQ_STRING":
            v = "u" + s.replace(u"\\\n", u"\\n")
        else:
            v = "u" + s
        return Value(ast.literal_eval(v), s, self.body)

    def boolean(self, b):
        """
        :meta private:
        """
        return Value(b.lower() in (u"true", u"yes", u"on"), b, self.body)

    def flatten(self, tree):
        """
        :meta private:
        """
        if isinstance(tree, Tree):
            for child in tree.children:
                for t in self.flatten(child):
                    yield t
        else:
            yield tree

    def list(self, *tokens):
        """
        :meta private:
        """
        return Value(
            [
                elt.value()
                for elt in self.flatten(tokens[1])
                if type(elt) is Value
            ],
            tokens,
            self.body,
        )

    def env_var(self, ev):
        """
        :meta private:
        """
        # TODO(olegs): We could also try to evaluate it, but this is
        # not important right now.
        return Value(ev, ev, self.body)

    def item(self, value):
        """
        :meta private:
        """
        return Item(value, self.body)

    def assignment(self, *tokens):
        """
        :meta private:
        """
        return Assignment(tokens, self.body)

    def function(self, *tokens):
        """
        :meta private:
        """
        return Function(tokens, self.body)

    def block(self, *tokens):
        """
        :meta private:
        """
        return Block(tokens, self.body)


class Writer(object):
    """Helper class intended to make writing the documents created by
    modifying the original libconfuse document easier.
    """

    def dump(self, conf, stream):
        """Writes the new contents of the document to the (file) stream.

        :param conf: The object to write.
        :type conf: ~cm_confuse.parser.ParseResult

        :param stream: (File) stream object, typically obtained by
            calling `open()` builtin.
        :type stream: io.TextIOBase
        """
        for token in conf.each():
            stream.write(str_type()(token))

    def dumps(self, conf):
        """Returns a string representation of `conf` object.  This is better
        than accessing the `body` of the `conf` object, as it may be
        in a "weird" state after modifictions (s.a. append or update).

        :param conf: The object to write.
        :type conf: ~cm_confuse.parser.ParseResult

        :returns: String representation of parse result.
        :rtype: str
        """
        buf = StringIO()
        self.dump(conf, buf)
        buf.seek(0)
        return buf.read()
